
/*
    unnc - decompresses nes files
    Copyright (C) 2005 Kuwanger

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    History
    v1.0 - September 18, 2005 - Based off gc, which was based off generic_cmp
                              - Now uses mmap, handles args, and tries to do
			        at least some sanity checks (no recompression)
			      - Only supports lzo compression, for the moment
			      - Extension npc stands for nes partial compression
    v1.1 - September 21, 2005 - Corrected history
    			      - Removed unused workspace
    v1.2 - September 23, 2005 - Make to work under Windows
    v1.3 - October 3, 2005    - Set more sane permissions (0600)
    */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <lzo1x.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define SI STDIN_FILENO
#define SO STDOUT_FILENO

#define LZO_COMPRESSION 0

struct ch {
	unsigned int size;
	unsigned int type; 
};

#define HEADER (sizeof(struct header))

#ifndef O_BINARY
#define O_BINARY 0
#endif

int usage(char *name)
{
	printf("Usage: %s <file.npc> <file.nes>\n", name);
	return 1;
}

int main (int argc, char *argv[])
{
  struct stat buf;
  char *src, *prg, *chr, *ptr;
  int in, out, prg_size, chr_size, temp;
  
  lzo_init ();

  if (argc == 3) {
	in = open(argv[1], O_RDONLY | O_BINARY);
	if (in < 0)
		return usage(argv[0]);
	fstat(in, &buf);
	src = malloc(buf.st_size);
	if (!src) {
		printf("Out of memory.\n");
		return 2;
	}
	if (buf.st_size != read(in, src, buf.st_size)) {
		printf("Error reading npc file.\n");
		return 3;
	}
	close(in);
	if (!(src[4]&0x80 || src[5]&0x80))
		return usage(argv[0]);
	prg_size = (src[4]&0x7f)*16384;
	chr_size = (src[5]&0x7f)*8192;
	ptr = &src[16];
	if (src[4]&0x80) {
		prg = malloc(prg_size);
		if (!prg) {
			printf("Out of memory.\n");
			return 2;
		}
		lzo1x_decompress(ptr+sizeof(struct ch), ((struct ch *) ptr)->size, prg, &temp, NULL);
		if (temp != prg_size) {
			printf("Error.  Decompressed PRG-ROM does not match expected size.\n");
			return 1;
		}
		ptr += ((struct ch *) ptr)->size;
	} else {
		prg = ptr;
		ptr += prg_size;
	}
	if (src[5]&0x80) {
		chr = malloc(chr_size);
		if (!chr) {
			printf("Out of memory.\n");
			return 2;
		}
		lzo1x_decompress(ptr+sizeof(struct ch), ((struct ch *) ptr)->size, chr, &temp, NULL);
		if (temp != chr_size) {
			printf("Error.  Decompressed CHR-ROM does not match expected size.\n");
			return 1;
		}
	} else {
		chr = ptr;
	}
	src[4] &= 0x7f;
	src[5] &= 0x7f;
	out = open(argv[2], O_CREAT | O_WRONLY | O_TRUNC | O_BINARY, 0600);
	if (out < 0)
		return usage(argv[0]);
	write(out, src, 16);
	write(out, prg, prg_size);
	write(out, chr, chr_size);
	close(out);
	printf("Outsize: %d\n", prg_size+chr_size+16);
	free(src);
  } else {
	return usage(argv[0]);
  }
  return 0; 
}

