#include <string.h>
#include "gba.h"
#include "minilzo.107/minilzo.h"
#include "cheat.h"
#include "rolimit_defines.h"

#define STATEID 0x57a731d7

#define STATESAVE 0
#define SRAMSAVE 1
#define CONFIGSAVE 2
#define CHEATLIST 3

#ifdef CHEATFINDER

extern u8 CHEATFINDER_CHEATS;
extern int num_cheats;
#endif

#define using_flashcart() on_flashcart

extern u8 on_flashcart;
extern u8 g_cartflags;	//(from iNES header)
extern char g_scaling;	//(cart.s) current display mode
extern char flicker;	//from ppu.s
extern u8 gammavalue;	//(ppu.s) current gammavalue
extern u8 stime;		//from ui.c
extern u8 autoA,autoB;	//from ui.c
extern u8 autostate;	//from ui.c
extern u8 checkmethod;  //from ui.c
extern u8 *textstart;	//from main.c

extern char pogoshell;	//main.c
extern int compressed; //main.c

int totalstatesize;		//how much SRAM is used

//-------------------
u8 *findrom(int);
u8 handle_rompages(u8*);
u8 handle_vrompages(u8*);
void cls(int);		//main.c
void drawtext(int,char*,int);
void setdarknessgs(int dark);
void scrolll(int f);
void scrollr(void);
void waitframe(void);
void sleeptimeconvert(void);
u32 getmenuinput(int);
void writeconfig(void);
void setup_sram_after_loadstate(void);
int titlecmp(char *, char *);

extern int roms;		//main.c
extern int selected;	//ui.c
extern romheader mb_header;
//----asm stuff------
int savestate(void*);		//cart.s
void loadstate(int,void*);		//cart.s

extern u8 *romstart;	//from cart.s
extern u32 romnum;	//from cart.s
extern u32 frametotal;	//from 6502.s
extern u32 joycfg;			//from io.s
//-------------------

int titlecmp(char *left, char *right)
{
	while (*left && *right && *left != '.' && *right != '.')
	{
		if (*left != *right)
			break;
		left++;
		right++;
	}
	return *right-*left;
}

void bytecopy(u8 *dst,u8 *src,int count) {
	int i=0;
	do {
		dst[i]=src[i];
		i++;
	} while(--count);
}

/*
void debug_(u32 n,int line);
void errmsg(char *s) {
	int i;

	drawtext(32+9,s,0);
	for(i=30;i;--i)
		waitframe();
	drawtext(32+9,"                     ",0);
}*/

void getsram() {		//copy GBA sram to WORKSPACE
	u8 *sram=MEM_SRAM;
	u8 *buff1=WORKSPACE;
	u32 *p;

	p=(u32*)buff1;
	if(*p!=STATEID) {	//if sram hasn't been copied already
		bytecopy(buff1,sram,0xe000);	//copy everything to WORKSPACE
		if(*p!=STATEID) {	//valid savestate data?
			*p=STATEID;	//nope.  initialize
			*(p+1)=0;
		}
	}
}

//quick & dirty rom checksum
u32 checksum(u8 *p) {
	u32 sum=0;
	int i;
	for(i=0;i<128;i++) {
		sum+=*p|(*(p+1)<<8)|(*(p+2)<<16)|(*(p+3)<<24);
		p+=128;
	}
	return sum;
}

u32 rom_reload() {
	u8 *p = findrom(romnum) + sizeof(romheader) + 16;

	handle_rompages(p);
	handle_vrompages(p);
}
	
u32 rom_checksum(int rom) {
	u8 *p = findrom(rom);

	if (p[sizeof(romheader)+4]&0x80 || p[sizeof(romheader)+5]&0x80)
		return *(u32 *) (p+sizeof(romheader)+8);
	
	return checksum(p+sizeof(romheader));
}

void writeerror() {
	int i;
	cls(2);
	drawtext(32+9, "  Write error! Memory full.",0);
	drawtext(32+10,"     Delete some saves.",0);
	for(i=90;i;--i)
		waitframe();
}

//(WORKSPACE=copy of GBA SRAM, newdata=new data)
//overwrite:  index=state#, erase=0
//new:  index=big number (anything >=total saves), erase=0
//erase:  index=state#, erase=1
//returns TRUE if successful
//IMPORTANT!!! totalstatesize is assumed to be current
int updatestates(int index,int erase,int type, u8 *data) {
	int i;
	int srcsize;
	int total=totalstatesize;
	u8 *src=WORKSPACE;
	u8 *dst;
	stateheader *newdata=(stateheader *)data;
	//stateheader *newdata=(stateheader*)BUFFER2;

	src+=4;//skip STATEID

	//skip ahead to where we want to write

	srcsize=((stateheader*)src)->size;
	i=(type==((stateheader*)src)->type)?0:-1;
	while(i<index && srcsize) {	//while (looking for state) && (not out of data)
		src+=srcsize;
		srcsize=((stateheader*)src)->size;
		if(((stateheader*)src)->type==type)
			i++;
	}

	dst=src;
	src+=srcsize;
	total-=srcsize;
	
	srcsize=((stateheader*)src)->size;
	
	//copy everything past i to buffer1
	while(srcsize) {
		memcpy(dst,src,srcsize);
		dst+=srcsize;
		src+=srcsize;
		srcsize=((stateheader*)src)->size;
	}
	
	// Append, if appropriate
	if(!erase) {
		srcsize = newdata->size;
		total += srcsize;
		if(total > 0xe000) //**OUT OF MEMORY**
			return 0;
		memcpy(dst, newdata, srcsize);
		dst += srcsize;
	}

	*(u32*)dst=0;	//terminate
	dst+=4;

	//copy everything to GBA sram

	totalstatesize=total;
	while(total<0xe000)
	{
		*dst++=0;
		total++;
	}
	bytecopy(MEM_SRAM,WORKSPACE,total);	//copy to sram
	return 1;
}

//more dumb stuff so we don't waste space by using sprintf
int twodigits(int n,char *s) {
	int mod=n%10;
	n=n/10;
	*(s++)=(n+'0');
	*s=(mod+'0');
	return n;
}

void getstatetimeandsize(char *s,int time,u32 size,u32 totalsize) {
	strcpy(s,"00:00:00 - 00/00k");
	twodigits(time/216000,s);
	s+=3;
	twodigits((time/3600)%60,s);
	s+=3;
	twodigits((time/60)%60,s);
	s+=5;
	twodigits(size/1024,s);
	s+=3;
	twodigits(totalsize/1024,s);
}

#define LOADMENU 0
#define SAVEMENU 1
#define SRAMMENU 2
#define FIRSTLINE 2
#define LASTLINE 16

//WORKSPACE holds copy of SRAM
//draw save/loadstate menu and update global totalstatesize
//returns a pointer to current selected state
//update *states on exit
stateheader* drawstates(int menutype,int *menuitems,int *menuoffset) {
	int type;
	int offset=*menuoffset;
	int sel=selected;
	int startline;
	int size;
	int statecount;
	int total;
	char s[30];
	stateheader *selectedstate;
	int time;
	int selectedstatesize;
	stateheader *sh=(stateheader*)(WORKSPACE+4);

	type=(menutype==SRAMMENU)?SRAMSAVE:STATESAVE;

	statecount=*menuitems;
	if(sel-offset>LASTLINE-FIRSTLINE-3 && statecount>LASTLINE-FIRSTLINE+1) {		//scroll down
		offset=sel-(LASTLINE-FIRSTLINE-3);
		if(offset>statecount-(LASTLINE-FIRSTLINE+1))	//hit bottom
			offset=statecount-(LASTLINE-FIRSTLINE+1);
	}
	if(sel-offset<3) {				//scroll up
		offset=sel-3;
		if(offset<0)					//hit top
			offset=0;
	}
	*menuoffset=offset;
	
	startline=FIRSTLINE-offset;
	cls(2);
	statecount=0;
	total=8;	//header+null terminator
	while(sh->size) {
		size=sh->size;
		if(sh->type==type) {
			if(startline+statecount>=FIRSTLINE && startline+statecount<=LASTLINE) {
				drawtext(32+startline+statecount,sh->title,sel==statecount);
			}
			if(sel==statecount) {		//keep info for selected state
				time=sh->framecount;
				selectedstatesize=size;
				selectedstate=sh;
			}
			statecount++;
		}
		total+=size;
		sh=(stateheader*)((u8*)sh+size);
	}

	if(sel!=statecount) {//not <NEW>
		getstatetimeandsize(s,time,selectedstatesize,total);
		drawtext(32+18,s,0);
	}
	if(statecount)
		drawtext(32+19,"Push SELECT to delete",0);
	if(menutype==SAVEMENU) {
		if(startline+statecount<=LASTLINE)
			drawtext(32+startline+statecount,"<NEW>",sel==statecount);
		drawtext(32,"Save state:",0);
		statecount++;	//include <NEW> as a menuitem
	} else if(menutype==LOADMENU) {
		drawtext(32,"Load state:",0);
	} else {
		drawtext(32,"Erase SRAM:",0);
	}
	*menuitems=statecount;
	totalstatesize=total;
	return selectedstate;
}

//compress src into dest (adding header), using 64k of workspace
void compressstate(u8 *dest, lzo_uint size,u16 type,u8 *src,void *workspace) {
	lzo_uint compressedsize;
	stateheader *sh;

/*
	if (workspace == NULL) {
		// No workspace, no compression.
		memcpy(dest+sizeof(stateheader),src,size);
		compressedsize=size;
	} else {
 */
	//workspace needs to be 64k
	lzo1x_1_compress(src,size,dest+sizeof(stateheader),&compressedsize,workspace);
	*(u32 *)WORKSPACE = 0; //Invalidate WORKSPACE as SRAM
/*	}
 */
	
	//setup header:
	sh=(stateheader*)dest;
	sh->size=(compressedsize+sizeof(stateheader)+3)&~3;	//size of compressed state+header, word aligned
	sh->type=type;
	sh->compressedsize=compressedsize;	//size of compressed state
	sh->framecount=frametotal;
	sh->checksum=rom_checksum(romnum);	//checksum
    if(pogoshell)
    {
		strcpy(sh->title,(char*)&(mb_header.name));
    }
    else
    {
		strcpy(sh->title,(char*)findrom(romnum));
    }
}

void managesram() {
	int i;
	int menuitems;
	int offset=0;

	getsram();

	selected=0;
	drawstates(SRAMMENU,&menuitems,&offset);
	if(!menuitems)
		return;		//nothing to do!

	scrolll(0);
	do {
		i=getmenuinput(menuitems);
		if(i&SELECT) {
			updatestates(selected,1,SRAMSAVE,NULL);
			if(selected==menuitems-1) selected--;	//deleted last entry.. move up one
		}
		if(i&(SELECT+UP+DOWN+LEFT+RIGHT))
			drawstates(SRAMMENU,&menuitems,&offset);
	} while(menuitems && !(i&(L_BTN+R_BTN+B_BTN)));
	scrollr();
}

void savestatemenu() {
	int i;
	int menuitems;
	int offset=0;

	i=savestate(RAWSTATE);
	compressstate(COMPRESSEDSTATE,i,STATESAVE,RAWSTATE,WORKSPACE);
	*(u32 *)WORKSPACE = 0; //Invalidate WORKSPACE as SRAM

	getsram();

	selected=0;
	drawstates(SAVEMENU,&menuitems,&offset);
	scrolll(0);
	do {
		i=getmenuinput(menuitems);
		if(i&(A_BTN)) {
			if(!updatestates(selected,0,STATESAVE,COMPRESSEDSTATE))
				writeerror();
		}
		if(i&SELECT)
			updatestates(selected,1,STATESAVE,NULL);
		if(i&(SELECT+UP+DOWN+LEFT+RIGHT))
			drawstates(SAVEMENU,&menuitems,&offset);
	} while(!(i&(L_BTN+R_BTN+A_BTN+B_BTN)));
	scrollr();
	rom_reload();
}

//locate last save by checksum
//returns save index (-1 if not found) and updates stateptr
//updates totalstatesize (so quicksave can use updatestates)
int findstate(u32 checksum,int type,stateheader **stateptr) {
	int state,size,foundstate,total;
	stateheader *sh;
	romheader *ch;

	getsram();
	sh=(stateheader*)(WORKSPACE+4);

	if (pogoshell)
		ch = &mb_header;
	else
		ch = (romheader *) findrom(romnum);

	state=-1;
	foundstate=-1;
	total=8;
	size=sh->size;
	while(size) {
		if(sh->type==type) {
			state++;
			if((checkmethod == 0 && sh->checksum==checksum) ||
			   (checkmethod == 1 && titlecmp(sh->title, ch->name) == 0)) {
				foundstate=state;
				*stateptr=sh;
			}
		}
		total+=size;
		sh=(stateheader*)(((u8*)sh)+size);
		size=sh->size;
	}
	totalstatesize=total;
	return foundstate;
}

u8 *out;
void uncompressstate(int rom,stateheader *sh) {
	lzo_uint statesize=sh->compressedsize;
	/* Uncompress either right after the compressed state (not overlap outside
	 * of WORKSPACE) or before it (possibly up to 7K (24K-(64K-24K-23K)) overlay.
	 * Either way the overlap doesn't extend over LOCALROM's 128K nor does it
	 * overlay anything non-replacable. */
	out = ((u8 *)sh)+sh->size;
	if (out + STATESIZE >= RAMEND)
		out = ((u8 *) sh) - STATESIZE;

	lzo1x_decompress((u8*)(sh+1),statesize,out,&statesize,NULL);
	loadstate(rom, out);
	*(u32 *)WORKSPACE = 0; //Invalidate WORKSPACE as SRAM
#if CHEATFINDER
	cheatload();
#endif
	frametotal=sh->framecount;		//restore global frame counter
	setup_sram_after_loadstate();	//handle sram packing
}

#ifdef CHEATFINDER
void cheatload() {
	stateheader *sh;
	int i;
	u16 *dst, *src;
	lzo_uint cheatsize;

	if(!using_flashcart())
		return;

	i=findstate(rom_checksum(romnum),CHEATLIST,&sh);
	if(i>=0) {
		/* The cheat list is in VRAM, so it's necessary to write
		 * in multiples of 2 */
		cheatsize=sh->compressedsize;
		lzo1x_decompress((u8*)(sh+1),cheatsize,RAWCHEAT,&cheatsize,NULL);
		src=(u16*)RAWCHEAT;
		dst=(u16*)(&CHEATFINDER_CHEATS);
		// Write up to an extra byte.
		for (i = 0; i < cheatsize/2+(cheatsize&1); i++)
		{
			*dst = *src;
			src++;
			dst++;
		}
		num_cheats=cheatsize/18;
	} else
		num_cheats=0;
}

void cheatsave() {
	stateheader *sh;
	int i;

	if(!using_flashcart())
		return;

	setdarknessgs(7);	//darken
	drawtext(32+9,"       Saving cheat.",0);

	i=findstate(rom_checksum(romnum),CHEATLIST,&sh);
	if(i<0) i=65536;	//make new save if one doesn't exist
	if (num_cheats) {
		compressstate(COMPRESSEDCHEAT,num_cheats*18,CHEATLIST,&CHEATFINDER_CHEATS,WORKSPACE);
		*(u32 *)WORKSPACE = 0; //Invalidate WORKSPACE as SRAM
		getsram();
		if(!updatestates(i,0,CHEATLIST,COMPRESSEDCHEAT))
			writeerror();
	} else if (i < 65536) {
		if(!updatestates(i,1,CHEATLIST,NULL))
			writeerror();
	}
	cls(2);
}
#endif

void quickload() {
	stateheader *sh;
	int i;

	if(!using_flashcart())
		return;

	i=findstate(rom_checksum(romnum),STATESAVE,&sh);
	if(i>=0)
		uncompressstate(romnum,sh);
}

void quicksave() {
	stateheader *sh;
	int i;

	if(!using_flashcart())
		return;

	setdarknessgs(7);	//darken
	drawtext(32+9,"           Saving.",0);

	i=savestate(RAWSTATE);
	compressstate(COMPRESSEDSTATE,i,STATESAVE,RAWSTATE,WORKSPACE);
	*(u32 *)WORKSPACE = 0; //Invalidate WORKSPACE as SRAM
	i=findstate(rom_checksum(romnum),STATESAVE,&sh);
	if(i<0) i=65536;	//make new save if one doesn't exist
	if(!updatestates(i,0,STATESAVE,COMPRESSEDSTATE))
		writeerror();
	rom_reload();
	cls(2);
}

void backup_nes_sram() {
	int i;
	configdata *cfg;
	stateheader *sh;
	lzo_uint compressedsize;

	if(!using_flashcart())
		return;

	i=findstate(0,CONFIGSAVE,(stateheader**)&cfg);	//find config
	if(i>=0 && cfg->sram_checksum) {	//SRAM is occupied?
		i=findstate(cfg->sram_checksum,SRAMSAVE,&sh);//find out where to save
		if(i>=0) {
			memcpy(COMPRESSEDSRAM,sh,sizeof(stateheader));//use old info, in case the rom for this sram is gone and we can't look up its name.
			lzo1x_1_compress(MEM_SRAM+0xe000,SRAMSIZE,COMPRESSEDSRAM+sizeof(stateheader),&compressedsize,WORKSPACE);	//workspace needs to be 64k
			*(u32 *)WORKSPACE = 0; //Invalidate WORKSPACE as SRAM
			sh=(stateheader*)COMPRESSEDSRAM;
			sh->size=(compressedsize+sizeof(stateheader)+3)&~3;	//size of compressed state+header, word aligned
			sh->compressedsize=compressedsize;	//size of compressed state
			getsram();
			updatestates(i,0,SRAMSAVE,COMPRESSEDSRAM);
			rom_reload();
		}
	}
}

//make new saved sram (using NES_SRAM contents)
//this is to ensure that we have all info for this rom and can save it even after this rom is removed
void save_new_sram() {
	compressstate(COMPRESSEDSRAM,SRAMSIZE,SRAMSAVE,NES_SRAM,WORKSPACE);
	*(u32 *)WORKSPACE = 0; //Invalidate WORKSPACE as SRAM
	getsram();
	updatestates(65536,0,SRAMSAVE,COMPRESSEDSRAM);
	rom_reload();
}

void get_saved_sram() {
	int i,j;
	u32 chk;
	configdata *cfg;
	stateheader *sh;
	lzo_uint statesize;

	if(!using_flashcart())
		return;

	if(g_cartflags&2) {	//if rom uses SRAM
		chk=rom_checksum(romnum);
		i=findstate(0,CONFIGSAVE,(stateheader**)&cfg);	//find config
		j=findstate(chk,SRAMSAVE,&sh);	//see if packed SRAM exists
		if(i>=0) if(chk==cfg->sram_checksum) {	//SRAM is already ours
			bytecopy(NES_SRAM,MEM_SRAM+0xe000,SRAMSIZE);
			if(j<0) save_new_sram();	//save it if we need to
			return;
		}
		if(j>=0) {//packed SRAM exists: unpack into NES_SRAM
			statesize=sh->compressedsize;
			lzo1x_decompress((u8*)(sh+1),statesize,NES_SRAM,&statesize,NULL);
		} else { //pack new sram and save it.
			save_new_sram();
		}
		bytecopy(MEM_SRAM+0xe000,NES_SRAM,SRAMSIZE);
		writeconfig();//register new sram owner
	}
}

void setup_sram_after_loadstate() {
	int i;
	u32 chk;
	configdata *cfg;

	if(g_cartflags&2) {	//if rom uses SRAM
		chk=rom_checksum(romnum);
		i=findstate(0,CONFIGSAVE,(stateheader**)&cfg);	//find config
		if(i>=0) if(chk!=cfg->sram_checksum) {//if someone else was using sram, save it
			backup_nes_sram();
		}
		bytecopy(MEM_SRAM+0xe000,NES_SRAM,SRAMSIZE);		//copy nes sram to real sram
		i=findstate(chk,SRAMSAVE,(stateheader**)&cfg);	//does packed SRAM for this rom exist?
		if(i<0)						//if not, create it
			save_new_sram();
		writeconfig();//register new sram owner
	}
}

void loadstatemenu() {
	stateheader *sh;
	u32 key;
	int i;
	int offset=0;
	int menuitems;
	u32 sum;

	getsram();

	selected=0;
	sh=drawstates(LOADMENU,&menuitems,&offset);
	if(!menuitems)
		return;		//nothing to load!

	scrolll(0);
	do {
		key=getmenuinput(menuitems);
		if(key&(A_BTN)) {
			sum=sh->checksum;
			i=0;
			do {
				if(sum==rom_checksum(i)) {	//find rom with matching checksum
					uncompressstate(i,sh);
					i=8192;
				}
				i++;
			} while(i<roms);
			if(i<8192) {
				// Restore g_rombase/romstart
				cls(2);
				drawtext(32+9,"       ROM not found.",0);
				for(i=0;i<60;i++)	//(1 second wait)
					waitframe();
			}
		} else if(key&SELECT) {
			updatestates(selected,1,STATESAVE,NULL);
			if(selected==menuitems-1) selected--;	//deleted last entry? move up one
		}
		if(key&(SELECT+UP+DOWN+LEFT+RIGHT))
			sh=drawstates(LOADMENU,&menuitems,&offset);
	} while(menuitems && !(key&(L_BTN+R_BTN+A_BTN+B_BTN)));
	scrollr();
}

const configdata configtemplate={
	sizeof(configdata),
	CONFIGSAVE,
	0,0,0,0,0,0,
	"CFG"
};

void writeconfig() {
	configdata *cfg;
	int i,j;

	if(!using_flashcart())
		return;

	i=findstate(0,CONFIGSAVE,(stateheader**)&cfg);
	if(i<0) {//make new config
		memcpy(RAWCONFIG,&configtemplate,sizeof(configdata));
		cfg=(configdata*)RAWCONFIG;
	}
	j =(g_scaling & 0xF);						//store current display type
	j |= (gammavalue & 0x7)<<5;					//store current gamma value
	cfg->displaytype = j;
	j = stime & 0x3;							//store current autosleep time
	j |= (autostate & 0x3)<<5;					//store current autostate setting
	j |= ((flicker & 0x1)^1)<<4;				//store current flicker setting
	cfg->misc = j;
	j = checkmethod & 0x1;
	j |= (autoA & 0x3) << 1;
	j |= (autoB & 0x3) << 3;
	j |= (joycfg & 0x400) >> 5;
	cfg->misc2 = j;
	if(g_cartflags&2) {							//update sram owner
			cfg->sram_checksum=rom_checksum(romnum);
	}
	if(i<0) {	//create new config
		updatestates(0,0,CONFIGSAVE,RAWCONFIG);
	} else {		//config already exists, update sram directly (faster)
		bytecopy((u8*)cfg-WORKSPACE+MEM_SRAM,(u8*)cfg,sizeof(configdata));
	}
}

void readconfig() {
	int i;
	configdata *cfg;
	if(!using_flashcart())
		return;

	i=findstate(0,CONFIGSAVE,(stateheader**)&cfg);
	if(i>=0) {
		i = cfg->displaytype;					//restore display type
		g_scaling = (i & 0xF);
		gammavalue = (i & 0xE0)>>5;				//restore gamma value
		i = cfg->misc;
		stime = i & 0x3;						//restore autosleep time
		sleeptimeconvert(); // Set actual sleeptime variable
		autostate = (i & 0x60)>>5;				//restore autostate setting
		flicker = ((i & 0x10)^0x10)>>4;			//restore flicker setting
		i = cfg->misc2;
		checkmethod = i & 0x1;
		autoA = (i & 0x6)>>1;
		autoB = (i & 0x18)>>3;
		joycfg |= (i & 0x60) << 5; //swapab
		joycfg|=A_BTN+(A_BTN<<16)+B_BTN+(B_BTN<<16);
		if(autoA==1)
			joycfg&=~A_BTN;
		else if(autoA==2)
			joycfg&=~(A_BTN<<16);
		if(autoB==1)
			joycfg&=~B_BTN;
		else if(autoA==2)
			joycfg&=~(B_BTN<<16);
	}
}

#if 0
void clean_nes_sram() {
	int i,j;
	u8 *nes_sram_ptr = MEM_SRAM+0xe000;
	configdata *cfg;

	if(!using_flashcart())
		return;

	for(i=0;i<SRAMSIZE;i++) *nes_sram_ptr++ = 0;

	i=findstate(0,CONFIGSAVE,(stateheader**)&cfg);
	if(i<0) {//make new config
		memcpy(RAWCONFIG,&configtemplate,sizeof(configdata));
		cfg=(configdata*)RAWCONFIG;
	}
	j = (g_scaling & 0xF);						//store current display type
	j |= (gammavalue & 0x7)<<5;					//store current gamma value
	cfg->displaytype = j;
	j = stime & 0xf;							//store current autosleep time
	j |= ((flicker & 0x1)^1)<<4;				//store current flicker setting
	cfg->misc = j;
	cfg->sram_checksum=0;						// we don't want to save the empty sram
	if(i<0) {	//create new config
		updatestates(0,0,CONFIGSAVE,RAWCONFIG);
	} else {		//config already exists, update sram directly (faster)
		bytecopy((u8*)cfg-WORKSPACE+MEM_SRAM,(u8*)cfg,sizeof(configdata));
	}
}
#endif

